#   Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class PassManager(object):
    """ pass管理器。
    """
    # pass_map存储name与其对应的pass
    pass_map = dict()

    def __init__(self):
        pass

    @staticmethod
    def add_new_pass(name, pass_):
        if name not in PassManager.pass_map:
            PassManager.pass_map[name] = pass_

    @staticmethod
    def clear():
        PassManager.passes = list()

    @staticmethod
    def lookup(name):
        return PassManager.pass_map[name]


def pass_register(cls):
    name = cls.get_name()
    PassManager.add_new_pass(name, cls)
    return cls
